#pragma once

#include <Game/Core.hpp>

#include <bt/behavior_node.hpp>

#include <chrono>
#include <fstream>
#include <queue>
#include <unordered_map>
#include <unordered_set>
#include <variant>

namespace std {
template <> struct hash<Tile> {
  std::size_t operator()(const Tile &tile) const {
    std::size_t h1 = std::hash<std::int32_t>()(tile.X);
    std::size_t h2 = std::hash<std::int32_t>()(tile.Y);
    std::size_t h3 = std::hash<std::int32_t>()(tile.Plane);
    return h1 ^ (h2 << 1) ^ (h3 << 2);
  }
};
} // namespace std

namespace bt::pathfinding {
class link {
public:
  Tile start;
  Tile destination;

  link(Tile start, Tile destination) noexcept
      : start(start), destination(destination) {}

  bool operator==(const link &other) const noexcept {
    return start == other.start;
  }

  bool operator!=(const link &other) const noexcept {
    return start != other.start;
  }

  virtual bool can_handle() const noexcept = 0;

  virtual bool handle() noexcept = 0;
};

using path_step = std::variant<Tile, std::shared_ptr<link>>;

using path = std::vector<path_step>;

class path_node {
public:
  path_step step;
  std::shared_ptr<path_node> parent;
  double cost;

  path_node(const path_step &step, std::shared_ptr<path_node> parent,
            double cost)
      : step(step), parent(parent), cost(cost) {}
};

class settings {
public:
  std::int32_t agility_level;

  settings() noexcept = default;  
};

std::unordered_set<std::int32_t> mapped_regions;
std::unordered_map<Tile, Pathfinding::COLLISION_FLAG> collision_data;
std::unordered_map<Tile, std::shared_ptr<link>> links;
 
void add_link(std::shared_ptr<link> link) noexcept {
  links.emplace(link->start, link);
}

void load_collision_data(std::ifstream &file) {
  std::string line;
  std::getline(file, line); // skip csv header
  while (std::getline(file, line)) {
    std::stringstream ss(line);
    std::string token;
    std::vector<std::string> tokens;
    while (std::getline(ss, token, ',')) {
      tokens.emplace_back(token);
    }

    const auto region = std::stoi(tokens[0]);
    const auto plane = std::stoi(tokens[1]);
    const auto x = std::stoi(tokens[2]);
    const auto y = std::stoi(tokens[3]);
    const auto flag =
        static_cast<Pathfinding::COLLISION_FLAG>(std::stoi(tokens[4]));

    mapped_regions.insert(region);
    collision_data.emplace(Tile(x, y, plane), flag);
  }
}

void visit_neighbors(const path_step &step, const settings& settings_, const auto &func) {
  Tile tile;
  if (std::holds_alternative<Tile>(step)) {
    tile = std::get<Tile>(step);
  } else {
    const auto &lnk = std::get<std::shared_ptr<link>>(step);
    tile = lnk->destination;
  }

  Pathfinding::COLLISION_FLAG flags = Pathfinding::COLLISION_FLAG::OPEN;
  auto it = collision_data.find(tile);
  if (it != collision_data.end()) {
    flags = it->second;
  }

  const auto blocked = [](Pathfinding::COLLISION_FLAG flags) {
    return (flags & Pathfinding::BLOCKED) || (flags & Pathfinding::OCCUPIED);
  };

  constexpr std::array<std::tuple<int, int, Pathfinding::COLLISION_FLAG>, 4>
      directions = {{{0, 1, Pathfinding::COLLISION_FLAG::NORTH},
                     {1, 0, Pathfinding::COLLISION_FLAG::EAST},
                     {0, -1, Pathfinding::COLLISION_FLAG::SOUTH},
                     {-1, 0, Pathfinding::COLLISION_FLAG::WEST}}};

  for (const auto &[dx, dy, dirFlag] : directions) {
    if ((flags & dirFlag) == Pathfinding::COLLISION_FLAG::OPEN) {
      int newX = tile.X + dx;
      int newY = tile.Y + dy;

      const auto neighbor = Tile(newX, newY, tile.Plane);

      const auto link_it = links.find(neighbor);
      if (link_it != links.end()) {
        func(link_it->second);
        continue;
      }

      auto neighbor_it = collision_data.find(neighbor);
      if (neighbor_it != collision_data.end()) {
        if (!blocked(neighbor_it->second)) {
          func(neighbor);
        }
      } else {
        func(neighbor);
      }
    }
  }
}

path find_path(const Tile &start, const Tile &end, const settings &settings_) {
  const auto compare = [](const std::shared_ptr<path_node> &a,
                          const std::shared_ptr<path_node> &b) {
    return a->cost > b->cost;
  };

  const auto heuristic = [](const path_step &a, const path_step &b) {
    const Tile &tileA =
        std::holds_alternative<Tile>(a)
            ? std::get<Tile>(a)
            : std::get<std::shared_ptr<link>>(a)->destination;
    const Tile &tileB =
        std::holds_alternative<Tile>(b)
            ? std::get<Tile>(b)
            : std::get<std::shared_ptr<link>>(b)->destination;

    return std::abs(tileA.X - tileB.X) + std::abs(tileA.Y - tileB.Y) +
           std::abs(tileA.Plane - tileB.Plane);
  };

  std::priority_queue<std::shared_ptr<path_node>,
                      std::vector<std::shared_ptr<path_node>>,
                      decltype(compare)>
      queue(compare);

  std::unordered_map<path_step, double> costs;
  std::unordered_set<path_step> visited;

  queue.emplace(std::make_shared<path_node>(start, nullptr, 0.0));
  costs[start] = 0.0;

  while (!queue.empty()) {
    auto current = queue.top();
    queue.pop();

    if (heuristic(current->step, end) == 0) { // Reached destination
      path p;
      while (current != nullptr) {
        p.emplace_back(current->step);
        current = current->parent;
      }

      std::reverse(p.begin(), p.end());
      return p;
    }

    visit_neighbors(current->step, settings_, [&](const auto &neighbor) {
      using neighbor_t = std::decay_t<decltype(neighbor)>;
      if constexpr (std::is_same_v<Tile, neighbor_t>) {
        double new_cost = costs[current->step] +
                          1; // Assuming constant cost of 1 for each step

        if (visited.find(neighbor) == visited.end() ||
            new_cost < costs[neighbor]) {
          costs[neighbor] = new_cost;
          double priority = new_cost + heuristic(neighbor, end);
          queue.emplace(
              std::make_shared<path_node>(neighbor, current, priority));
          visited.insert(neighbor);
        }
      } else if constexpr (std::is_same_v<std::shared_ptr<link>, neighbor_t>) {
        auto new_cost = costs[current->step] + 1;
        if (visited.find(neighbor->destination) == visited.end() ||
            new_cost < costs[neighbor->destination]) {
          costs[neighbor->destination] = new_cost;
          double priority = new_cost + heuristic(neighbor->destination, end);
          queue.emplace(
              std::make_shared<path_node>(neighbor, current, priority));
          visited.insert(neighbor);
        }
      }
    });
  }

  return path();
}

void paint_path(Tile start, Tile end, const settings &settings_) {
  auto path_ = find_path(start, end, settings_);
  for (const auto &step : path_) {
    if (const auto tile = std::get_if<Tile>(&step)) {
      Paint::DrawTile(*tile, 255, 0, 0, 255);
    }
  }
}
} // namespace bt::pathfinding